"""
Provides `UnitSymbol` an immutable class representing base unit symbols.
"""

from typing import NamedTuple
import re


class UnitSymbol(
    NamedTuple("_UnitSymbol", symbol=str, name=str, quantity=str),
):
    """
    Immutable class inheriting from `NamedTuple` representing a
    base unit symbol with fields:
    * symbol: The symbol displayed in unit expressions. Should be a valid
      Python identifier.
    * name: The unit name.
    * quantity: The physical quantity represented by the unit.

    ``` python
    m = UnitSymbol('m', 'meter', 'length')
    s = UnitSymbol(symbol='s', name='second', quantity='time')
    ```
    """

    __slots__ = ()
    __reg_expr = "^[A-Za-z_][A-Za-z0-9_]*"

    def __new__(cls, symbol: str, name: str, quantity: str):
        if not UnitSymbol.is_valid_symbol(symbol):
            raise ValueError(
                f"Symbol {type(symbol)}:{symbol} is not a valid identifier."
            )

        return super().__new__(cls, symbol, name, quantity)

    @classmethod
    def is_valid_symbol(cls, value: str) -> bool:
        """
        Returns `True` if `value` matches the regular expression:
        '^[A-Za-z_][A-Za-z0-9_]*'
        """
        if not isinstance(value, str):
            return False

        if re.fullmatch(UnitSymbol.__reg_expr, value) is None:
            return False

        return True



"""
Numpy array with the additional attribute `unit`.
"""

from __future__ import annotations

from typing import Callable, Union

import numpy as np
from numpy.core._exceptions import UFuncTypeError

from .errors import OperationNotSupported
from .quantity import Quantity
from .unit import UnitBase, UnitExprBase


class qarray(np.ndarray):
    """
    An array with elements representing a quantity that can be
    described by a numerical value and a unit.

    `qarray` is a sub-class of ndarray with the additional instance
    attribute `unit`.

    Implementation closely follows:
    https://numpy.org/devdocs/user/basics.subclassing.html#basics-subclassing
    """

    _unit_types = (UnitBase, UnitExprBase)
    __slots__ = ("__unit",)

    def __new__(
        subtype,
        shape,
        dtype=float,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        unit=1.0,
    ):
        # The call in the next line triggers a call to
        # qarray.__array_finalize__
        obj = super().__new__(
            subtype, shape, dtype, buffer, offset, strides, order
        )
        obj.unit = unit
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(qarray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. qarray():
        #    obj is None
        #    (we're in the middle of the qarray.__new__
        #    constructor, and self.unit will be set when we return to
        #    qarray.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(qarray):
        #    obj is arr
        #    (type(obj) can be qarray)
        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is qarray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'unit', because this
        # method sees all creation of default objects - with the
        # qarray.__new__ constructor, but also with
        # arr.view(qarray).
        self.__unit = getattr(obj, "unit", 1.0)

    @classmethod
    def from_input(cls, input, unit=1.0) -> qarray:
        """Constructs a `qarray` from an existing ndarray
        or from a (nested) sequence of entries.
        """
        obj = np.asarray(input).view(cls)
        obj.unit = unit
        return obj

    @property
    def unit(self):
        """Returns the unit of the object."""
        return self.__unit

    @unit.setter
    def unit(self, value) -> None:
        factor = (
            value.factor
            if isinstance(value, self._unit_types)
            else float(value)
        )

        if factor == 1.0:
            self.__unit = value
            return None

        if factor == 0.0:
            raise ValueError(
                f"Could not set unit with zero magnitude: {value}."
            )

        try:
            self *= factor
            self.__unit = value / factor
        except UFuncTypeError:
            cfactor = self.dtype.type(factor)
            if cfactor == factor:
                self *= cfactor
                self.__unit = value / factor
            else:
                # If factor can not be safely converted to dtype do not
                # normalize unit:
                self.__unit = value

    def __str__(self) -> str:
        return super().__str__() + " unit: " + str(self.unit)

    def __repr__(self) -> str:
        return super().__repr__()[:-1] + f", unit={self.unit})"

    def __add__(self, other) -> qarray:
        other_unit = getattr(other, "unit", 1.0)
        # If units match simply add arrays.
        if self.unit == other_unit:
            return super().__add__(other)

        alpha = other_unit / self.unit

        if isinstance(alpha, float):
            return super().__add__(other * alpha)

        if (
            isinstance(alpha, UnitExprBase)
            and alpha.base_exponents == alpha.base_exponents_zero
        ):
            return super().__add__(other * alpha.base_factor)

        raise OperationNotSupported(self, other, "+")

    def __sub__(self, other) -> qarray:
        other_unit = getattr(other, "unit", 1.0)

        # If units match simply subtract arrays.
        if self.unit == other_unit:
            return super().__sub__(other)

        alpha = other_unit / self.unit

        if isinstance(alpha, float):
            return super().__sub__(other * alpha)

        if (
            isinstance(alpha, UnitExprBase)
            and alpha.base_exponents == alpha.base_exponents_zero
        ):
            return super().__sub__(other * alpha.base_factor)

        raise OperationNotSupported(self, other, "-")

    def __mul__(self, other) -> qarray:
        """
        Returns the result of multiplying the united ndarray `self`
        with `other`.
        """
        if isinstance(other, self._unit_types):
            obj = self.copy()
            obj.unit = self.unit * other
            return obj

        if isinstance(other, Quantity):
            obj = super().__mul__(other.value)
            obj.unit = self.unit * other.unit
            return obj

        obj = super().__mul__(other)
        other_unit = getattr(other, "unit", 1.0)
        obj.unit = self.unit * other_unit
        return obj

    def __rmul__(self, other) -> qarray:
        """
        Returns the result of multiplying `other` with the united array
        `self`.
        """
        if isinstance(other, self._unit_types):
            obj = self.copy()
            obj.unit = other * self.unit
            return obj

        if isinstance(other, Quantity):
            obj = super().__rmul__(other.value)
            obj.unit = other.unit * self.unit
            return obj

        obj = super().__rmul__(other)
        other_unit = getattr(other, "unit", 1.0)
        obj.unit = other_unit * self.unit
        return obj

    def __truediv__(self, other) -> qarray:
        """
        Returns the result of dividing the united ndarray `self`
        with `other`.
        """
        if isinstance(other, self._unit_types):
            obj = self.copy()
            obj.unit = self.unit / other
            return obj

        if isinstance(other, Quantity):
            obj = super().__truediv__(other.value)
            obj.unit = self.unit / other.unit
            return obj

        obj = super().__truediv__(other)
        other_unit = getattr(other, "unit", 1.0)
        obj.unit = self.unit / other_unit
        return obj

    def __rtruediv__(self, other) -> qarray:
        """
        Returns the result of dividing `other` by the united ndarray `self`.
        """
        if isinstance(other, self._unit_types):
            obj = super().__rtruediv__(other.factor)
            obj.unit = other / self.unit
            return obj

        if isinstance(other, Quantity):
            obj = super().__rtruediv__(other.value)
            obj.unit = other.unit / self.unit
            return obj

        obj = super().__rtruediv__(other)
        other_unit = getattr(other, "unit", 1.0)
        obj.unit = other_unit / self.unit
        return obj

    def __pow__(self, other: Union[float, int]) -> qarray:
        obj = super().__pow__(other)
        obj.unit = self.unit.__pow__(other)
        return obj

    def __abs__(self) -> qarray:
        obj = super().__abs__()
        obj.unit = self.unit
        return obj

    def __neg__(self) -> qarray:
        obj = super().__neg__()
        obj.unit = self.unit
        return obj

    def __pos__(self) -> qarray:
        obj = super().__pos__()
        obj.unit = self.unit.__pos__()
        return obj

    def __eq__(self, other) -> qarray:
        result = self.compare(other, super().__eq__)
        if result is None:
            result = super().__eq__(other)
            result.unit = 1.0
            result.fill(False)
        return result

    def __le__(self, other) -> qarray:
        result = self.compare(other, super().__le__)
        if result is None:
            raise OperationNotSupported(self, other, "<=")
        return result

    def __ge__(self, other) -> qarray:
        result = self.compare(other, super().__ge__)
        if result is None:
            raise OperationNotSupported(self, other, ">=")
        return result

    def __gt__(self, other) -> qarray:
        result = self.compare(other, super().__gt__)
        if result is None:
            raise OperationNotSupported(self, other, ">")
        return result

    def __lt__(self, other) -> qarray:
        result = self.compare(other, super().__lt__)
        if result is None:
            raise OperationNotSupported(self, other, "<")
        return result

    def compare(self, other, comparison_operator: Callable) -> qarray:
        """
        Generic comparison function that handles input of type `Number`,
        `ndarray` and `qarray` using `comparison_operator`.

        Returns `None` if the comparison failed due to incompatible units.
        """
        other_unit = getattr(other, "unit", 1.0)

        if self.unit == other_unit:
            obj = comparison_operator(other)
            obj.unit = 1.0
            return obj

        alpha = other_unit / self.unit

        if isinstance(alpha, float):
            obj = comparison_operator(other * alpha)
            obj.unit = 1.0
            return obj

        if (
            isinstance(alpha, UnitExprBase)
            and alpha.base_exponents == alpha.base_exponents_zero
        ):
            obj = comparison_operator(other * alpha.base_factor)
            obj.unit = 1.0
            return obj

        return None



"""
Si Unit System based on meter, second, kilogram, Ampere, Kelvin, mole, candela.
"""

from math import pi

from .unit import UnitBase, UnitMeta
from .unit_symbol import UnitSymbol

unit_symbols = (
    UnitSymbol("m", "meter", "length"),
    UnitSymbol("s", "second", "time"),
    UnitSymbol("kg", "kilogram", "mass"),
    UnitSymbol("A", "ampere", "electric current"),
    UnitSymbol("K", "kelvin", "temperature"),
    UnitSymbol("mol", "mole", "amount of substance"),
    UnitSymbol("cd", "candela", "luminous intensity"),
)


class SiUnit(UnitBase, metaclass=UnitMeta, unit_symbols=unit_symbols):
    """
    Class representing SI unitexpr. The fundamental units are available as:
    `SiUnit.m`, `SiUnit.s`, `SiUnit.kg`, `SiUnit.A`, `SiUnit.K`, `SiUnit.mol`
    and `SiUnit.cd`.
    """

    __slots__ = ()


# Base units
m = SiUnit.m
"""
Base unit: Meter
"""
s = SiUnit.s
kg = SiUnit.kg
A = SiUnit.A
K = SiUnit.K
mol = SiUnit.mol
cd = SiUnit.cd


# Derived units
au = SiUnit("au", "astronomical unit", "length", expr=149597870700 * m)
sr = SiUnit("sr", "steradian", "solid angle", expr=SiUnit.expr_type.one)
N = SiUnit("N", "newton", "force", expr=kg * m * s ** -2)
J = SiUnit("J", "joule", "energy", expr=N * m)
eV = SiUnit("eV", "electronvolt", "energy", expr=1.602176634e-19 * J)
W = SiUnit("W", "watt", "power", expr=J / s)
C = SiUnit("C", "coulomb", "charge", expr=A * s)
V = SiUnit("V", "volt", "electric potential difference", expr=J / C)
F = SiUnit("F", "farad", "capacitance", expr=C / V)
ohm = SiUnit("ohm", "ohm", "resistance", expr=V / A)
Pa = SiUnit("Pa", "pascal", "pressure", expr=N * m ** -2)
S = SiUnit("S", "siemens", "electrical conductance", 1 / ohm)
Wb = SiUnit("Wb", "weber", "magnetic flux", expr=V * s)
T = SiUnit("T", "tesla", "magnetic flux density", expr=Wb * m ** -2)
H = SiUnit("H", "henry", "inductance", expr=Wb / A)
lm = SiUnit("lm", "lumen", "luminous flux", expr=cd / sr)
Bq = SiUnit("Bq", "bequerel", "radioactivity (decays per second)", expr=1 / s)
lx = SiUnit("lx", "lux", "illuminance", expr=lm * m ** -2)
Gy = SiUnit("Gy", "gray", "absorbed dose of ionizing radiation", expr=J / kg)

# Constants with exact value
delta_nu_Cs = SiUnit(
    "delta_nu_cs",
    "hyperfine transition of frequency of Cs",
    "frequency",
    expr=91926377 / s,
)

c = SiUnit("c", "speed of light", "velocity", expr=299792458 * m / s)
h = SiUnit("h", "Planck constant", "angular momentum", 6.62607015e-34 * J * s)
h_bar = SiUnit(
    "h_bar", "Reduced Planck constant", "angular momentum", h / (2 * pi)
)
e = SiUnit("e", "elementary charge", "charge", expr=1.606176634e-19 * C)
k = SiUnit(
    "k", "Boltzmann constant", "energy/Kelvin", expr=1.380649e-23 * J / K
)

m_e = SiUnit("m_e", "electron mass", "weigth", expr=9.1093837015e-31 * kg)

N_a = SiUnit(
    "N_a",
    "Avogadro constant",
    "number of molecules per mole",
    expr=6.02214076e-23 / mol,
)

K_cd = SiUnit(
    "K_cd",
    "luminous efficacy",
    "luminous efficacy of 540 THz radiation",
    expr=683 * lm / W,
)



"""
Module providing custom error classes.
"""


class OperationNotSupported(Exception):
    """
    Error raised if an arithmetic operation involving unit expressions
    is not supported.

    Attributes:
        - `left`: Left operand.
        - `right`: Right operand.
        - `operator`: Arithmetic operator used in the expression.
        - `message`: Explanation of the error.
    """

    # pylint: disable=too-many-function-args
    def __init__(self, left, right, operator, addon_message=""):
        self.left = left
        self.right = right
        self.operator = operator
        self.addon_message = addon_message
        super().__init__(self)

    def __str__(self):
        """
        Returns the error message string.
        """
        return (
            "Could not evaluate: "
            + f"{self.left.__repr__()} {self.operator}"
            + f" {self.right.__repr__()} ."
            + self.addon_message
        )

    def __repr__(self):
        """
        Return the error message string.
        """
        return self.__str__(self)

    def invert(self):
        """
        Changes the order of the operands. (left, right) = (right, left)
        """
        (self.left, self.right) = (self.right, self.left)
        return self



"""
Units, unit expressions and united numpy arrays.

The package `unitexpr` provides the modules:

- unit: Module containing the base classes `UnitBase` and `UnitMeta`.
- unit_symbol: Provides the class `UnitSymbol` used to specify
  base unit symbols.
- qarray: A sub-class of a `numpy` n-dimensional array whose entries
  represent a quantity with an associated unit.
- si_units: Predefined units representing the SI Unit System.
- sc_units: Predefined units based on (nm, ps, m_e, K, cd, ). Semiconductor
            Units.


"""

from .unit import UnitBase, UnitMeta
from .unit_symbol import UnitSymbol
from .qarray import qarray
from .quantity import Quantity



"""
Semiconductor Unit System based on nanometer, picosecond,
electron mass, Ampere,
Kelvin, mol, and candela.
"""

from math import pi

from .unit import UnitBase, UnitMeta
from .unit_symbol import UnitSymbol

unit_symbols = (
    UnitSymbol("nm", "nano-meter", "length"),
    UnitSymbol("ps", "pico-second", "time"),
    UnitSymbol("m_e", "electron rest mass", "mass"),
    UnitSymbol("A", "ampere", "electric current"),
    UnitSymbol("K", "kelvin", "temperature"),
    UnitSymbol("mol", "mole", "amount of substance"),
    UnitSymbol("cd", "candela", "luminous intensity"),
)


class ScUnit(UnitBase, metaclass=UnitMeta, unit_symbols=unit_symbols):
    """
    Class representing Semiconductor unitexpr. The fundamental units
    are available as:
    `ScUnit.nm`, `ScUnit.ps`, `ScUnit.m_e`, `ScUnit.A`, `ScUnit.K`,
    `ScUnit.mol`, and `ScUnit.cd`.
    """


# Base units
nm = ScUnit.nm
ps = ScUnit.ps
m_e = ScUnit.m_e
A = ScUnit.A
K = ScUnit.K
mol = ScUnit.mol
cd = ScUnit.cd

# Base units in terms of SI units:
m = ScUnit("m", "meter", "length", expr=1.0e9 * nm)
s = ScUnit("s", "second", "time", expr=1.0e12 * ps)
kg = ScUnit("kg", "kilogram", "mass", expr=(1.0 / 9.1093837015) * 1e31 * m_e)

# Derived units
au = ScUnit("au", "astronomical unit", "length", expr=149597870700 * m)
sr = ScUnit("sr", "steradian", "solid angle", expr=ScUnit.expr_type.one)
N = ScUnit("N", "newton", "force", expr=kg * m * s ** -2)
J = ScUnit("J", "joule", "energy", expr=N * m)
eV = ScUnit("eV", "electronvolt", "energy", expr=1.602176634e-19 * J)
W = ScUnit("W", "watt", "power", expr=J / s)
C = ScUnit("C", "coulomb", "charge", expr=A * s)
V = ScUnit("V", "volt", "electric potential difference", expr=J / C)
F = ScUnit("F", "farad", "capacitance", expr=C / V)
ohm = ScUnit("ohm", "ohm", "resistance", expr=V / A)
Pa = ScUnit("Pa", "pascal", "pressure", expr=N * m ** -2)
S = ScUnit("S", "siemens", "electrical conductance", 1 / ohm)
Wb = ScUnit("Wb", "weber", "magnetic flux", expr=V * s)
T = ScUnit("T", "tesla", "magnetic flux density", expr=Wb * m ** -2)
H = ScUnit("H", "henry", "inductance", expr=Wb / A)
lm = ScUnit("lm", "lumen", "luminous flux", expr=cd / sr)
Bq = ScUnit("Bq", "bequerel", "radioactivity (decays per second)", expr=1 / s)
lx = ScUnit("lx", "lux", "illuminance", expr=lm * m ** -2)
Gy = ScUnit("Gy", "gray", "absorbed dose of ionizing radiation", expr=J / kg)

# Constants with exact value
delta_nu_Cs = ScUnit(
    "delta_nu_cs",
    "hyperfine transition of frequency of Cs",
    "frequency",
    expr=91926377 / s,
)

c = ScUnit("c", "speed of light", "velocity", expr=299792458 * m / s)
h = ScUnit("h", "Planck constant", "angular momentum", 6.62607015e-34 * J * s)
h_bar = ScUnit(
    "h_bar", "Reduced Planck constant", "angular momentum", h / (2 * pi)
)
e = ScUnit("e", "elementary charge", "charge", expr=1.606176634e-19 * C)
k = ScUnit(
    "k", "Boltzmann constant", "energy/Kelvin", expr=1.380649e-23 * J / K
)

m_e = ScUnit("m_e", "electron mass", "weigth", expr=9.1093837015e-31 * kg)

N_a = ScUnit(
    "N_a",
    "Avogadro constant",
    "number of molecules per mole",
    expr=6.02214076e-23 / mol,
)

K_cd = ScUnit(
    "K_cd",
    "luminous efficacy",
    "luminous efficacy of 540 THz radiation",
    expr=683 * lm / W,
)
delta_nu_Cs = ScUnit(
    "delta_nu_cs",
    "hyperfine transition of frequency of Cs",
    "frequency",
    expr=91926377 * 1.0e-12 / ps,
)



"""
Provides the classes `UnitExprMeta` and `UnitExprBase`.
"""

from __future__ import annotations

from operator import add, neg, sub
from typing import TYPE_CHECKING, Any, Iterable, NamedTuple, Tuple, Type, Union

from lockattrs import protect
from .errors import OperationNotSupported
from .unit_dict import UnitDict
from .unit_symbol import UnitSymbol

if TYPE_CHECKING:
    from .unit import UnitBase


class UnitExprMeta(type):
    """
    Meta-class of objects representing unit expressions. Initializes
    class attributes.
    """

    def __new__(
        cls,
        cls_name,
        bases,
        attrs,
        unit_symbols: Tuple[UnitSymbol],
        unit_type: Type,
    ):
        attrs["unit_symbols"] = tuple(unit_symbols)
        attrs["dim"] = len(unit_symbols)
        attrs["base_exponents_zero"] = (0.0,) * attrs["dim"]
        attrs["unit_type"] = unit_type
        return super().__new__(cls, cls_name, bases, attrs)

    def __init__(
        self,
        cls_name,
        bases,
        attrs,
        unit_symbols: Iterable[UnitSymbol],
        unit_type: Type,
    ):
        self.one = self(
            terms=(),
            exponents=(),
            factor=1.0,
            base_exponents=self.base_exponents_zero,
            base_factor=1.0,
        )
        self.valid_types = (self, self.unit_type)
        self.expr_type = self
        super().__init__(cls_name, bases, attrs)

    @protect()
    def __setattr__(self, name: str, value: Any) -> None:
        return super().__setattr__(name, value)


class UnitExprMixin:

    __slots__ = ()

    def __or__(self, other) -> UnitExprBase:
        if isinstance(other, self.valid_types):
            if other.base_exponents != self.base_exponents:
                raise OperationNotSupported(self, other, "|")
            return

        if isinstance(other, (int, float)):
            if self.base_exponents != self.base_exponents_zero:
                return False
            return True
        return False

    def proportional_to(self, other) -> bool:
        """
        Returns `True` if the unit
        `self` can be converted to to the unit expression or unit `other`
        using a constant factor.

        Returns `False` otherwise.
        """
        if isinstance(other, self.valid_types):
            if other.base_exponents != self.base_exponents:
                return False
            return True

        if isinstance(other, (int, float)):
            if self.base_exponents != self.base_exponents_zero:
                return False
            return True
        return False

    def scaling_factor(self, other) -> Union[float, None]:
        """
        Returns the scaling factor that converts the unit
        `self` to the unit expression or unit `other`.

        Returns `None` if `self` cannot be converted to `other`
        by multiplication with a number of type `int` or `float`.
        """
        if isinstance(other, self.valid_types):
            if other.base_exponents != self.base_exponents:
                return None
            return other.base_factor / self.base_factor

        if isinstance(other, (int, float)):
            if self.base_exponents != self.base_exponents_zero:
                return None
            return other / self.base_factor
        return None

    def common_factors(self, other) -> Union[Tuple[float, float], None]:
        """
        Returns the tuple `(self.base_factor, other.base_factor)`
        if self and other have the same `base_exponents` otherwise `None`.

        Returns the tuple `(self.base_factor, other)` if `self` is number-like
        that is all entries of `self.base_exponents` are zero and other is
        of type  int` or `float`.
        """
        if isinstance(other, self.valid_types):
            if other.base_exponents != self.base_exponents:
                return None
            return (other.base_factor, self.base_factor)

        if isinstance(other, (int, float)):
            if self.base_exponents != self.base_exponents_zero:
                return None
            return (other, self.base_factor)
        return None

    def __abs__(self) -> UnitExprBase:
        """
        Returns a unit expression representing the absolute value of `self`.
        """
        return self.expr_type(
            terms=self.terms,
            exponents=self.exponents,
            factor=abs(self.factor),
            base_exponents=self.base_exponents,
            base_factor=abs(self.base_factor),
        )

    def __add__(self, other) -> UnitExprBase:
        """
        Returns the result of adding the unit expression `self` and `other`.
        """
        if (
            isinstance(other, self.valid_types)
            and self.base_exponents == other.base_exponents
        ):
            if self.base_factor == 0 or not self.terms:
                return (
                    other
                    if isinstance(other, UnitExprBase)
                    else other.self_expr
                )

            base_factor = self.base_factor + other.base_factor
            factor = base_factor / self.base_factor * self.factor

            return self.expr_type(
                self.terms,
                self.exponents,
                factor,
                self.base_exponents,
                base_factor,
            )

        if (
            isinstance(other, (float, int))
            and self.base_exponents == self.base_exponents_zero
        ):
            base_factor = self.base_factor + other

            if self.base_factor == 0.0 or not self.terms:
                return self.one * base_factor

            factor = base_factor / self.base_factor * self.factor

            return self.expr_type(
                self.terms,
                self.exponents,
                factor,
                self.base_exponents,
                base_factor,
            )

        raise OperationNotSupported(self, other, "+")

    def __radd__(self, other) -> UnitExprBase:
        """
        Returns the result of adding `other` and the unit expression `self`.
        """
        if (
            isinstance(other, self.valid_types)
            and self.base_exponents == other.base_exponents
        ):

            if other.base_factor == 0.0 or not other.terms:
                return self

            base_factor = other.base_factor + self.base_factor
            factor = base_factor / other.base_factor * other.factor

            return self.expr_type(
                other.terms,
                other.exponents,
                factor,
                other.base_exponents,
                base_factor,
            )

        if (
            isinstance(other, (float, int))
            and self.base_exponents == self.base_exponents_zero
        ):
            if other == 0:
                return self

            if self.base_factor == 0.0 or not self.terms:
                return self.one * other

            base_factor = other + self.base_factor
            factor = base_factor / self.base_factor * self.factor

            return self.expr_type(
                self.terms,
                self.exponents,
                factor,
                self.base_exponents,
                base_factor,
            )

        raise OperationNotSupported(other, self, "+")

    def __pos__(self):
        return self

    def __sub__(self, other):
        """
        Returns the result of subtracting `other` from `self`.
        """
        if (
            isinstance(other, self.valid_types)
            and self.base_exponents == other.base_exponents
        ):
            if self.base_factor == 0.0 or not self.terms:
                return -other

            base_factor = self.base_factor - other.base_factor
            factor = base_factor / self.base_factor * self.factor

            return self.expr_type(
                self.terms,
                self.exponents,
                factor,
                self.base_exponents,
                base_factor,
            )

        if (
            isinstance(other, (float, int))
            and self.base_exponents == self.base_exponents_zero
        ):
            base_factor = self.base_factor - other

            if self.base_factor == 0.0 or not self.terms:
                return self.one * base_factor

            factor = base_factor / self.base_factor * self.factor

            return self.expr_type(
                self.terms,
                self.exponents,
                factor,
                self.base_exponents,
                base_factor,
            )

        raise OperationNotSupported(self, other, "-")

    def __rsub__(self, other):
        """
        Returns the result of subtracting `self` from `other`.
        """
        if (
            isinstance(other, self.valid_types)
            and self.base_exponents == other.base_exponents
        ):
            if other.base_factor == 0.0 or not other.terms:
                return -self

            base_factor = other.base_factor - self.base_factor
            factor = base_factor / other.base_factor * other.factor

            return self.expr_type(
                other.terms,
                other.exponents.exponents,
                factor,
                other.base_exponents,
                base_factor,
            )

        if (
            isinstance(other, (float, int))
            and self.base_exponents == self.base_exponents_zero
        ):
            base_factor = other - self.base_factor

            if self.base_factor == 0.0 or not self.terms:
                return self.one * base_factor

            factor = base_factor / self.base_factor * self.factor

            return self.expr_type(
                self.terms,
                self.exponents,
                factor,
                self.base_exponents,
                base_factor,
            )

        raise OperationNotSupported(other, self, "-")

    def __neg__(self) -> UnitExprBase:
        """
        Negation operator.
        """
        return self.expr_type(
            self.terms,
            self.exponents,
            factor=-self.factor,
            base_exponents=self.base_exponents,
            base_factor=-self.base_factor,
        )

    def __mul__(self, other) -> "UnitExprBase":
        """
        Returns the result of multiplying `self` with `other`.

        - `self`: Left operand.
        - `other`: Right operand`.
        """
        if isinstance(other, self.valid_types):
            return self.expr_type(
                self.terms + other.terms,
                self.exponents + other.exponents,
                self.factor * other.factor,
                tuple(map(add, self.base_exponents, other.base_exponents)),
                self.base_factor * other.base_factor,
            )

        if isinstance(other, (int, float)):
            return self.expr_type(
                self.terms,
                self.exponents,
                self.factor * other,
                self.base_exponents,
                self.base_factor * other,
            )

        try:
            return other.__rmul__(self)
        except TypeError as error:
            raise OperationNotSupported(self, other, "*") from error

    def __rmul__(self, other) -> "UnitExprBase":
        """
        Returns the result of multiplying `other` with `self`.

        - `other`: Left operand.
        - `self`: Right operand`.
        """
        if isinstance(other, self.valid_types):
            return self.expr_type(
                other.terms + self.terms,
                other.exponents + self.exponents,
                other.factor * self.factor,
                tuple(map(add, self.base_exponents, other.base_exponents)),
                self.base_factor * other.base_factor,
            )

        if isinstance(other, (int, float)):
            return self.expr_type(
                self.terms,
                self.exponents,
                other * self.factor,
                self.base_exponents,
                other * self.base_factor,
            )
        raise OperationNotSupported(other, self, "*")

    def __truediv__(self, other) -> "UnitExprBase":
        """
        Returns the result of: `self` / `other`.
        - `self`: Left operand.
        - `other`: Right operand.
        """
        if isinstance(other, (int, float)):
            return self.expr_type(
                self.terms,
                self.exponents,
                self.factor / other,
                self.base_exponents,
                self.base_factor / other,
            )

        if isinstance(other, self.valid_types):
            return self.expr_type(
                self.terms + other.terms,
                self.exponents + tuple(map(neg, other.exponents)),
                self.factor / other.factor,
                tuple(map(sub, self.base_exponents, other.base_exponents)),
                self.base_factor / other.base_factor,
            )
        # Handles e.g. objects of type qarray divided by unit/unit_expr.
        try:
            return other.__rtruediv__(self)
        except TypeError:
            raise OperationNotSupported(self, other, "/")

    def __rtruediv__(self, other) -> UnitExprBase:
        """
        Returns the result of: `other` / `self`.
        - `self`: Right operand.
        - `other`: Left operand.
        """
        if isinstance(other, (int, float)):
            return self.expr_type(
                self.terms,
                tuple(map(neg, self.exponents)),
                other / self.factor,
                tuple(-exponent for exponent in self.base_exponents),
                other / self.base_factor,
            )

        if isinstance(other, self.valid_types):
            return self.expr_type(
                other.terms + self.terms,
                tuple(map(neg, other.exponents)) + self.exponents,
                other.factor / self.factor,
                tuple(map(sub, other.base_exponents, self.base_exponents)),
                other.base_factor / self.base_factor,
            )

        raise OperationNotSupported(other, self, "/")

    def __pow__(self, other: int) -> UnitExprBase:
        """
        Returns the result of: self**other.
        """
        if isinstance(other, (int, float)):
            return self.expr_type(
                self.terms,
                tuple([exponent * other for exponent in self.exponents]),
                self.factor ** other,
                tuple([exponent * other for exponent in self.base_exponents]),
                self.base_factor ** other,
            )

        raise OperationNotSupported(self, other, "**")

    def __le__(self, other: Union[UnitBase, UnitExprBase, int, float]) -> bool:
        """Returns `self.base_factor <= other.base_factor`.

        Raises and error of type `OperationNotSupported` if the unit
        expressions are not comparable.
        """

        factors = self.common_factors(other)
        if factors is None:
            raise OperationNotSupported(self, other, "<=")
        left, right = factors
        return True if left <= right else False

    def __ge__(self, other: Union[UnitBase, UnitExprBase, int, float]) -> bool:
        """Returns `self.base_factor >= other.base_factor`.

        Raises and error of type `OperationNotSupported` if the unit
        expressions are not comparable.
        """
        factors = self.common_factors(other)
        if factors is None:
            raise OperationNotSupported(self, other, ">=")
        left, right = factors
        return True if left >= right else False

    def __lt__(self, other: Union[UnitBase, UnitExprBase, int, float]) -> bool:
        """Returns `self.base_factor < other.base_factor`.

        Raises and error of type `OperationNotSupported` if the unit
        expressions are not comparable.
        """
        factors = self.common_factors(other)
        if factors is None:
            raise OperationNotSupported(self, other, "<")
        left, right = factors
        return True if left < right else False

    def __gt__(self, other: Union[UnitBase, UnitExprBase, int, float]) -> bool:
        """
        Returns `self.base_factor > other.base_factor`.

        Raises and error of type `OperationNotSupported` if the unit
        expressions are not comparable.
        """
        factors = self.common_factors(other)
        if factors is None:
            raise OperationNotSupported(self, other, ">")
        left, right = factors
        return True if left > right else False


class UnitExprBase(
    NamedTuple(
        "_UnitExprBase",
        terms=Tuple["UnitBase"],
        exponents=Tuple[float],
        factor=float,
        base_exponents=Tuple[float],
        base_factor=float,
    ),
):
    """
    Immutable class representing an expression (multiplication) of
    units and real numbers.
    """

    __slots__ = ()

    def __new__(
        cls,
        terms: Tuple[UnitBase],
        exponents: Tuple[float],
        factor: float,
        base_exponents: Tuple[float],
        base_factor: float,
    ):
        """
        Constructs a new unit expression instance.
        This method is reserved for internal use. Parameters are not
        validated.

        To create consistent unit expression combine
        existing units and unit expressions using the arithmetic operators
        `+`, `-`, `*`, `/`, and `**`.

        * terms: A tuple with entries of type `cls.unit_type`.
        * exponents: The expression exponent for each term.
        * factor: A scaling factor of type `int` or `float`.
        * base_exponents: The exponents with respect to the system
        base units.
        * base_factor: The scaling factor of the expression when written
        in terms of base unit.
        """
        if cls is UnitExprBase:
            raise TypeError(f"Instantiation of {cls} not allowed.")

        return super().__new__(
            cls, terms, exponents, factor, base_exponents, base_factor
        )

    @classmethod
    def from_dict(cls, dexpr: dict, factor: float = 1.0):
        """
        Returns a unit expression representing the terms and
        exponents specified in `dict` and scaled with `factor`.

        Note: Unit expression classes are available after creating
        a unit class and can be accessed using the
        class attribute `expr_type` (see example below).

        ``` python
        # Type shortcut
        SiUnitExpr = SiUnit.expr_type
        expr = SiUnitExpr(dexpr={m:1, s:-1}, factor=10.0)
        print(expr)
        # Prints: 10.0*m*s**-1
        ```
        """
        base_exponents = cls.base_exponents_zero
        base_factor = factor
        for term, exponent in dexpr.items():
            cls.validate_term(term)
            base_exponents = list(
                map(
                    add,
                    base_exponents,
                    [x * exponent for x in term.base_exponents],
                )
            )
            base_factor *= term.base_factor ** exponent

        return cls(
            tuple(dexpr.keys()),
            tuple(dexpr.values()),
            factor,
            tuple(base_exponents),
            base_factor,
        )

    @classmethod
    def validate_term(cls, term) -> None:
        """
        Raises a `TypeError` if `term` does not have type `cls.unit_type`.
        """
        if not isinstance(term, cls.unit_type):
            raise TypeError(
                "Unit expression contains an invalid term: "
                + f"'{term}' of type '{type(term)}'. \n "
                + f"Expected objects of type: {cls}."
            )

    @property
    def dexpr(self):
        """
        Return the unit expression in the form of a dictionary with keys
        containing the unit symbols and values containing the
        corresponding exponent.

        Note: The actual expression is scaled with the instance attribute
        `factor`.
        """
        dexpr = UnitDict()
        for term, exponent in zip(self.terms, self.exponents):
            dexpr[term] = dexpr.get(term, 0.0) + exponent
        return dexpr.filter_value(0.0)

    def __str__(self):
        """
        Returns a string representing the expression defined by `terms`
        and `factor`.
        """
        out = ""
        for term, exponent in self.dexpr.items():
            if exponent == 1.0:
                out = out + "*" + str(term)
            elif exponent == 0.0:
                continue
            else:
                out = out + "*" + str(term) + "**" + str(exponent)

        if not out:
            return str(self.factor)

        # Strip leading '*'
        out = out[1:]
        if self.factor != 1.0:
            out = str(self.factor) + "*" + out
        return out

    def __repr__(self):
        """
        Returns a string represenation of self.
        """

        out = ""
        for term, exponent in self.dexpr.items():
            if exponent == 1.0:
                out = out + "*" + str(term)
            elif exponent == 0.0:
                continue
            else:
                out = out + "*" + str(term) + "**" + str(exponent)

        if not out:
            return str(self.factor) + "*" + self.__class__.__name__ + ".one"

        # Strip leading '*'
        out = out[1:]
        if self.factor != 1.0:
            out = str(self.factor) + "*" + out
        return out

    @property
    def repr(self) -> NamedTuple:
        """
        Returns a namedtuple containing detailed object information.
        """
        return super().__repr__()

    @property
    def base_repr(self):
        """
        Returns a string representing the unit expression in terms
        of base units.
        """
        out = ""
        for term, exponent in zip(self.unit_symbols, self.base_exponents):
            if exponent == 1.0:
                out = out + "*" + term.symbol
            elif exponent == 0.0:
                continue
            else:
                out = out + "*" + term.symbol + "**" + str(exponent)
        out = out[1:]
        if not out:
            out = str(self.base_factor)
        elif self.base_factor != 1.0:
            out = str(self.base_factor) + "*" + out
        return out

    @property
    def base_expr(self):
        """
        Returns an expression of self in terms of base units.
        """
        dexpr = {}
        for term, exponent in zip(
            self.unit_type.base_units, self.base_exponents
        ):
            if exponent == 0.0:
                continue
            dexpr[term] = exponent

        return self.__class__(
            terms=dexpr.keys(),
            exponents=dexpr.values(),
            factor=self.base_factor,
            base_exponents=self.base_exponents,
            base_factor=self.base_factor,
        )

    def __eq__(self, other: object) -> bool:
        """
        Returns `True` if:
        * the expressions `self` and `other` match when
          resolved in terms of base units.
        * other is numeric and the resolved expression of `self` represents the
        same number. (All base exponents must be zero).
        """
        if isinstance(other, self.valid_types):
            return (
                self.base_factor == other.base_factor
                and self.base_exponents == other.base_exponents
            )

        if isinstance(other, (int, float)):
            return (
                self.base_exponents == self.base_exponents_zero
                and self.base_factor == other
            )

        return False

    def __ne__(self, other: object) -> bool:
        """
        Returns `True` if:
        * the expressions `self` and `other` do not match when
          resolved in terms of base units.
        * other is numeric and the resolved expression of `self` represents a
        different number. (All base exponents must be zero).
        """
        if isinstance(other, self.valid_types):
            return (
                self.base_factor != other.base_factor
                or self.base_exponents != other.base_exponents
            )

        if isinstance(other, (int, float)):
            return (
                self.base_exponents != self.base_exponents_zero
                or self.base_factor != other
            )

        return True

    def __req__(self, other: object) -> bool:
        """
        Returns `True` if:
        * the expressions `self` and `other` match when
          resolved in terms of base units.
        * other is numeric and the resolved expression of `self` represents the
        same number.
        """
        return self.__eq__(other)

    def scaling_factor(self, other) -> Union[float, None]:
        """
        Returns the scaling factor that converts the unit expression
        `self` to the unit expression or unit `other`.

        Returns `None` if `self` cannot be converted to `other`
        by multiplication with a number of type `int` or `float`.
        """
        if isinstance(other, self.valid_types):
            if other.base_exponents != self.base_exponents:
                return None
            return other.base_factor / self.base_factor

        if isinstance(other, (int, float)):
            if self.base_exponents != self.base_exponents_zero:
                return None
            return other / self.base_factor

        return None



"""
Module providing the classes:

* UnitMeta
* UnitBase
"""
from __future__ import annotations

from typing import Any, Iterable, NamedTuple, Tuple

from lockattrs import protect

from ._unitexpr import UnitExprBase, UnitExprMeta, UnitExprMixin
from .unit_symbol import UnitSymbol


class UnitMeta(type):
    """
    Meta class of objects representing physical units.
    """

    def __new__(
        cls, cls_name, bases, attrs, unit_symbols: Iterable[UnitSymbol]
    ):
        attrs["__slots__"] = tuple()

        return super().__new__(
            cls,
            cls_name,
            (UnitExprMixin, *bases),
            attrs,
        )

    def __init__(
        self, cls_name, bases, attrs, unit_symbols: Iterable[UnitSymbol]
    ):

        # Define the unit expression type.
        expr_type = UnitExprMeta(
            cls_name=cls_name + "Expr",
            bases=(UnitExprMixin, UnitExprBase),
            attrs={"__slots__": tuple()},
            unit_symbols=unit_symbols,
            unit_type=self,
        )
        self.expr_type = expr_type

        # Define base units.
        base_dimensions = len(unit_symbols)
        zero = [0.0] * base_dimensions
        base_units = []
        for index, unit_symbol in enumerate(unit_symbols):
            vector = zero.copy()
            vector[index] = 1.0
            base_unit = self(
                unit_symbol.symbol,
                unit_symbol.name,
                unit_symbol.quantity,
                self.expr_type(
                    terms=(unit_symbol.symbol,),
                    exponents=(1.0,),
                    factor=1.0,
                    base_exponents=tuple(vector),
                    base_factor=1.0,
                ),
            )
            base_units.append(base_unit)
            setattr(self, unit_symbol.symbol, base_unit)

        self.base_units = tuple(base_units)
        self.base_exponents_zero = tuple(zero)
        self.valid_types = (self, self.expr_type)
        self.info_type = NamedTuple(
            "UnitInfo",
            symbol=str,
            name=str,
            quantity=str,
            terms=Tuple[UnitMeta],
            exponents=Tuple[float],
            factor=float,
            base_exponents=Tuple[float],
            base_factor=float,
            sub_terms=Tuple[str],
            sub_exponents=Tuple[str],
            sub_factor=float,
        )
        super().__init__(cls_name, bases, attrs)

    def is_base_unit(self, unit):
        """
        Returns `True` if `unit` is a base unit and `False` otherwise.
        """
        return unit in self.base_units

    # The decorator protects all attributes from modification
    # (after they have been initially set).
    # Any attempt at modification will raise an error of type
    # ProtectedAttributeError.
    @protect()
    def __setattr__(self, name: str, value: Any) -> None:
        return super().__setattr__(name, value)


class UnitBase(
    NamedTuple(
        "_UnitBase",
        symbol=str,
        name=str,
        quantity=str,
        base_exponents=Tuple[float],
        base_factor=float,
        sub_terms=Tuple[UnitMeta],
        sub_exponents=Tuple[float],
        sub_factor=float,
    )
):
    """Base class of objects representing physical units.

    To generate unit systems subclass `UnitBase` providing the base unit
    symbols. The base units will be available as class attributes (see
    example below).

    ``` python
    # Defining unit symbols
    unit_symbols = (
        UnitSymbol(symbol='m','name'='meter',quantity='length'),
        UnitSymbol(symbol='s','name'='second',quantity='time'),
    )
    # Sub-classing the base class `UnitBase`
    class MetricUnit(UnitBase, metaclass=UnitMeta,
          unit_symbols=unit_symbols):
        pass
    # Base units are available as class attributes.
    m = MetricUnit.m
    s = MetricUnit.s
    # Declaring derived units
    c = MetricUnit('c', 'speed of light', 'velocity', expr=299792458*m/s)
    ```
    """

    __slots__ = ()

    def __new__(
        cls, symbol: str, name: str, quantity: str, expr: UnitExprBase
    ):
        if cls == UnitBase:
            raise TypeError(f"Class {cls} must be subclassed.")

        if not isinstance(expr, cls.expr_type):
            raise TypeError(
                f"Expected expression of type {cls.expr_type}. "
                + f"Found {type(expr)}."
            )

        return super().__new__(
            cls,
            symbol,
            name,
            quantity,
            base_exponents=expr.base_exponents,
            base_factor=expr.base_factor,
            sub_terms=expr.terms,
            sub_exponents=expr.exponents,
            sub_factor=expr.factor,
        )

    def __eq__(self, other: object) -> bool:
        """
        Returns `True` if:
        * the unit `self` and the unit/unit-expression `other` match when
        resolved in terms of base units.
        * other is numeric and the resolved expression of `self` represents a
        the same number. (All base exponents must be zero).
        """
        if isinstance(other, self.expr_type):
            return (
                self.base_factor == other.base_factor
                and self.base_exponents == other.base_exponents
            )

        if isinstance(other, (int, float)):
            return (
                self.base_exponents == self.base_exponents_zero
                and self.base_factor == other
            )

        return self is other

    def __ne__(self, other: object) -> bool:
        """
        Returns `True` if:
        * the unit `self` and the unit/unit-expression `other`
          do not match when resolved in terms of base units.
        * other is numeric and the resolved expression of `self` represents a
        different number. (All base exponents must be zero).
        """
        if isinstance(other, self.expr_type):
            return (
                self.base_factor != other.base_factor
                or self.base_exponents != other.base_exponents
            )

        if isinstance(other, (int, float)):
            return (
                self.base_exponents != self.base_exponents_zero
                or self.base_factor != other
            )

        return self is not other

    def __hash__(self) -> int:
        return id(self)

    def __repr__(self):
        """
        Returns a string representation of the unit object.
        """
        return self.symbol

    def __str__(self):
        """
        Returns a string representation of the unit object.
        """
        return self.symbol

    @property
    def info(self) -> NamedTuple:
        """
        Returns a `NamedTuple` containing detailed object information.
        """
        return self.info_type(
            self.symbol,
            self.name,
            self.quantity,
            (self,),
            self.exponents,
            self.factor,
            self.base_exponents,
            self.base_factor,
            self.sub_terms,
            self.sub_exponents,
            self.sub_factor,
        )

    @property
    def expr(self) -> UnitExprBase:
        """
        Returns an expression representing `self` in terms of
        `sub_terms`.
        ``` python
        J = SiUnit('J', 'Joule', 'energy', expr=N*m)
        print(J.expr)  # prints: N*m
        ```
        """
        return self.expr_type(
            terms=self.sub_terms,
            exponents=self.sub_exponents,
            factor=self.sub_factor,
            base_exponents=self.base_exponents,
            base_factor=self.base_factor,
        )

    @property
    def self_expr(self) -> UnitExprBase:
        """
        Returns an expression representing `self` in terms of `self`.
        An alternative method of converting a unit to a unit expression is
        multiplication by 1.0.
        ``` python
        J = SiUnit('J', 'Joule', 'energy', expr=N*m)
        assert J.self_expr == J
        assert J.self_expr == 1.0*J
        assert type(J.self_expr) == SiUnit.expr_type

        print(J_expr) # Prints: J
        ```
        """
        return self.expr_type(
            terms=(self,),
            exponents=(1.0,),
            factor=1.0,
            base_exponents=self.base_exponents,
            base_factor=self.base_factor,
        )

    @property
    def factor(self) -> float:
        """
        Returns the (scaling) `factor` of the unit expression.
        Always returns 1.0.
        """
        return 1.0

    @property
    def terms(self) -> Tuple[str]:
        """
        Returns the `terms` of the unit expression.
        """
        return (self,)

    @property
    def exponents(self) -> Tuple[float]:
        """
        Returns the exponents of the unit expression.
        """
        return (1.0,)

    @property
    def base_expr(self) -> UnitExprBase:
        """
        Returns an expression representing self in terms of base units.
        """
        dexpr = {}
        for term, exponent in zip(self.base_units, self.base_exponents):
            if exponent == 0.0:
                continue
            dexpr[term] = exponent
        return self.expr_type(
            dexpr.keys(),
            dexpr.values(),
            factor=self.base_factor,
            base_exponents=self.base_exponents,
            base_factor=self.base_factor,
        )

    # ---------
    # Operators
    # ---------



"""
Dictionary intended to be used with numeric value types.
Adds support for arithmetic operators.
"""
from __future__ import annotations


class UnitDict(dict):
    """
    Extends `dict` adding support for the arithmetic
    operators `+`, `-`, scalar multiplication.

    * The value type must implement the operators above.
    * Strips keys with value zero.
    """

    # def __init__(self, *args, **kwargs):
    #    super().__init__(float, *args, **kwargs)

    def __sub__(self: UnitDict, other: UnitDict) -> UnitDict:
        """
        Subtracts the values of entries with the same key andipyt
        returns the resulting
        dictionary. Keys with value zero are deleted.

        Note: Returns a new object.
        """
        obj = dict(self)
        for term in other:
            tmp = obj[term] = obj.get(term, 0.0) - other[term]
            if not tmp:
                del obj[term]
        return self.__class__(obj)

    def copy(self) -> UnitDict:
        """
        Returns a new `UnitDict` object containing the same entries as `self`.
        """
        return self.__class__(self)

    def __add__(self: UnitDict, other: UnitDict) -> UnitDict:
        """
        Adds the values of entries with the same key and returns the resulting
        dictionary. Keys with value zero are deleted.

        Note: Returns a new object.
        """
        obj = UnitDict(self)
        for term in other:
            tmp = obj[term] = obj.get(term, 0.0) + other[term]
            if not tmp:
                del obj[term]
        return obj

    def __mul__(self: UnitDict, other: float) -> UnitDict:
        """
        Multiplies the value of each entry with `other` and returns the
        resulting dictionary.

        Note: Returns a new object.
        """
        if other == 0.0:
            return UnitDict()

        obj = UnitDict(self)
        for key in obj:
            obj[key] *= other
        return obj

    def __rmul__(self, other: float) -> UnitDict:
        """
        Multiplies `other` with the value of each entry and returns the
        resulting dictionary.

        Note: Returns a new object.
        """
        return self.__mul__(other)

    def __repr__(self) -> str:
        """
        Returns a string representation of self.
        """
        return self.__class__.__name__ + f"({super().__repr__()})"

    def __neg__(self) -> UnitDict:
        """
        Negates the dictionary values and returns the resulting
        dictionary.

        Note: Returns a new object.
        """
        obj = UnitDict(self)
        for key in obj:
            obj[key] *= -1
        return obj

    def filter_value(self, value) -> UnitDict:
        """
        Remove all entries with matching `value` and returns the resulting
        dictionary.

        Note: Returns a new object.
        """
        obj = UnitDict(self)
        for key in self.keys():
            if obj[key] == value:
                del obj[key]
        return obj


class ImmutableUnitDict(UnitDict):
    """
    Immutable class that extends `UnitDict` with support for the arithmetic
    operators `+`, `-`, scalar multiplication, and
    `**`.

    Note: The value type must implement the operators above.
    """

    def __add__(self: ImmutableUnitDict, other: UnitDict) -> ImmutableUnitDict:
        """
        Adds the values of entries with the same key and returns the resulting
        dictionary. Keys with value zero are deleted.

        Note: Returns a new object.
        """
        obj = dict(self)
        for term in other:
            tmp = obj[term] = obj.get(term, 0.0) + other[term]
            if not tmp:
                del obj[term]
        return ImmutableUnitDict(obj)

    def __sub__(self: ImmutableUnitDict, other: UnitDict) -> UnitDict:
        """
        Subtracts the values of entries with the same key and
        returns the resulting
        dictionary. Keys with value zero are deleted.

        Note: Returns a new object.
        """
        obj = dict(self)
        for term in other:
            tmp = obj[term] = obj.get(term, 0.0) - other[term]
            if not tmp:
                del obj[term]
        return ImmutableUnitDict(obj)

    def __rmul__(self, other: float) -> UnitDict:
        """
        Multiplies `other` with the value of each entry and returns the
        resulting dictionary.

        Note: Returns a new object.
        """
        return self.__mul__(other)

    def __mul__(self, other: float) -> ImmutableUnitDict:
        """
        Multiplies the value of each entry with `other` and returns the
        resulting dictionary.

        Note: Returns a new object.
        """
        if other == 0.0:
            return ImmutableUnitDict()

        obj = dict(self)
        for key in obj:
            obj[key] *= other
        return ImmutableUnitDict(obj)

    def __neg__(self) -> ImmutableUnitDict:
        obj = dict(self)
        for key in obj:
            obj[key] *= -1
        return ImmutableUnitDict(obj)

    def filter_value(self, value) -> ImmutableUnitDict:
        """
        Removes all entries with matching `value` and returns the resulting
        dictionary.

        Note: Returns a new object.
        """
        obj = dict(self)
        for key in self:
            if obj[key] == value:
                del obj[key]
        return ImmutableUnitDict(obj)

    def __setitem__(self, key, value) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object does not "
            + "support item assignment."
        )

    def __delitem__(self, value) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object does not "
            + "support item deletion."
        )

    def clear(self) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object " + "cannot be cleared."
        )

    def update(self, *args, **kwargs) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object " + "cannot be updated."
        )

    def setdefault(self, key, default) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object does not "
            + "support default values."
        )

    def pop(self) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object does not " + "support `pop`."
        )

    def popitem(self) -> None:
        """
        Not supported. Immutable object.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object does not "
            + "support `popitem`."
        )



from __future__ import annotations

from numbers import Number
from typing import Callable, Union

from .errors import OperationNotSupported
from .unit import UnitBase, UnitExprBase


class Quantity:
    """
    A quantity that can be described by a numerical value
    of type `int`, `float`, or `complex` and a unit.
    The attribute `info` can be used to store object information in the
    form of a string.
    """

    _unit_types = (UnitBase, UnitExprBase)

    __slots__ = ("__value", "__unit", "info")

    def __new__(
        cls,
        value: Number,
        unit: Union[UnitBase, UnitExprBase, float] = 1.0,
        info: str = "",
    ) -> Quantity:
        """
        Returns an object of type `Quantity`. The object attributes
        are set in `__init__`.
        """
        return super().__new__(cls)

    def __init__(
        self,
        value: Number,
        unit: Union[UnitBase, UnitExprBase, float] = 1.0,
        info: str = "",
    ):
        """
        Initializes object attributes.
        """
        self.info = info
        self.__value, self.__unit = self.__normalize(value, unit)
        self.__unit = unit

    @classmethod
    def __idem(
        cls,
        value: Number,
        unit: Union[UnitBase, UnitExprBase, float] = 1.0,
        info: str = "",
    ) -> Quantity:
        """
        Constructs an object of type `Quantity` without normalizing the unit.
        """
        obj = super().__new__(cls)
        obj.__value = value
        obj.__unit = unit
        obj.info = info
        return obj

    @property
    def unit(self):
        """Returns the unit of the quantity."""
        return self.__unit

    @property
    def value(self):
        """Returns the numerical value of the quantity."""
        return self.__value

    @classmethod
    def __normalize(cls, value, unit) -> tuple:
        """
        Returns a tuple with entries scaled value and normalized unit.
        The function is used in `__init__`.
        """
        if not isinstance(value, Number):
            raise TypeError(
                f"Parameter 'value' must be numerical. Found {value}."
            )

        try:
            factor = (
                unit.factor
                if isinstance(unit, cls._unit_types)
                else float(unit)
            )
        except ValueError as error:
            raise ValueError(
                "Expected a 'unit', a 'unit expression', or input that be "
                + f"converted to a float. Found:{unit.__repr__()}."
            ) from error

        if factor == 1.0:
            return (value, unit)

        if factor == 0.0:
            raise ValueError(
                f"Could not set unit with zero magnitude: {value}."
            )
        return (value * factor, unit / factor)

    def __str__(self) -> str:
        """Returns a string representing `self`."""
        return f"{self.__value} {self.__unit}"

    def __repr__(self) -> str:
        """Returns a string representing `self`."""
        result = (
            f"{self.__class__.__name__}({self.__value}, "
            + f"unit={self.__unit}"
        )

        if self.info.strip():
            result += f", info={self.info.__repr__()})"
        else:
            result += ")"
        return result

    def copy(self):
        """Returns a copy of `self`."""
        return self.__idem(self.__value, self.__unit, self.info)

    def __add__(self, other) -> Quantity:
        """Returns the result of adding `self` and `other`."""
        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)

        # If units match simply add values.
        if self.__unit == other_unit:
            return self.__idem(self.__value + other_value, self.__unit)

        alpha = other_unit / self.__unit

        if isinstance(alpha, float):
            return self.__idem(self.__value + other_value * alpha, self.__unit)

        if (
            isinstance(alpha, UnitExprBase)
            and alpha.base_exponents == alpha.base_exponents_zero
        ):
            return self.__idem(
                self.__value + other_value * alpha.base_factor, self.__unit
            )
        raise OperationNotSupported(self, other, "+")

    def __sub__(self, other) -> Quantity:
        """Returns the result of subtracting `other` from `self`."""
        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)

        # If units match simply subtract values.
        if self.__unit == other_unit:
            return self.__idem(self.__value - other_value, self.__unit)

        alpha = other_unit / self.__unit

        if isinstance(alpha, float):
            return self.__idem(self.__value - other_value * alpha, self.__unit)

        if (
            isinstance(alpha, UnitExprBase)
            and alpha.base_exponents == alpha.base_exponents_zero
        ):
            return self.__idem(
                self.__value - other_value * alpha.base_factor, self.__unit
            )
        raise OperationNotSupported(self, other, "-")

    def __mul__(self, other) -> Quantity:
        """
        Returns the result of multiplying `self` with `other`.
        """
        if isinstance(other, self._unit_types):
            return self.__class__(self.__value, self.__unit * other)

        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)
        try:
            return self.__class__(
                self.__value * other_value, self.__unit * other_unit
            )
        except TypeError:
            return NotImplemented

    def __rmul__(self, other) -> Quantity:
        """
        Returns the result of multiplying `other` with `self`.
        """
        if isinstance(other, self._unit_types):
            return self.__class__(self.__value, other * self.__unit)

        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)
        return self.__class__(
            other_value * self.__value,
            other_unit * self.__unit,
        )

    def __truediv__(self, other) -> Quantity:
        """
        Returns the result of dividing the `self` by `other`.
        """
        if isinstance(other, self._unit_types):
            return self.__class__(self.__value, self.__unit / other)

        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)

        try:
            return self.__class__(
                self.__value / other_value,
                self.__unit / other_unit,
            )
        except TypeError:
            return NotImplemented

    def __rtruediv__(self, other) -> Quantity:
        """
        Returns the result of dividing `other` by `self`.
        """
        if isinstance(other, self._unit_types):
            return self.__class__(self.__value, other / self.__unit)

        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)
        return self.__class__(
            other_value / self.__value,
            other_unit / self.__unit,
        )

    def __pow__(self, other: Union[float, int]) -> Quantity:
        """Returns the result of `self ** other`."""
        return self.__class__(self.__value ** other, self.__unit ** other)

    def __abs__(self) -> Quantity:
        """Returns `abs(self)`. Note: The operator affects the value only."""
        return self.__idem(self.__value.__abs__(), self.__unit)

    def __neg__(self) -> Quantity:
        """Returns `-self`. Note: The operator affects the value only."""
        return self.__idem(-self.__value, self.__unit)

    def __pos__(self) -> Quantity:
        """Returns `+self`. Note: The operator affects the value only."""
        return self.__idem(self.__value.__pos__(), self.__unit)

    def __eq__(self, other) -> bool:
        """Returns `True` if `self == other` and `False` otherwise."""
        result = self.compare(other, self.__value.__eq__)
        if result is None:
            return False
        return result

    def __le__(self, other) -> bool:
        """Returns `True` if `self <= other` and `False` otherwise."""
        result = self.compare(other, self.__value.__le__)
        if result is None:
            raise OperationNotSupported(self, other, "<=")
        return result

    def __ge__(self, other) -> bool:
        """Returns `True` if `self >= other` and `False` otherwise."""
        result = self.compare(other, self.__value.__ge__)
        if result is None:
            raise OperationNotSupported(self, other, ">=")
        return result

    def __gt__(self, other) -> bool:
        """Returns `True` if `self > other` and `False` otherwise."""
        result = self.compare(other, self.__value.__gt__)
        if result is None:
            raise OperationNotSupported(self, other, ">")
        return result

    def __lt__(self, other) -> bool:
        """Returns `True` if `self < other` and `False` otherwise."""
        result = self.compare(other, self.__value.__lt__)
        if result is None:
            raise OperationNotSupported(self, other, "<")
        return result

    def compare(
        self, other, comparison_operator: Callable
    ) -> Union[bool, None]:
        """
        Generic comparison function that handles input of type `Number`,
        `ndarray` and `qarray` using `comparison_operator`.

        Returns `None` if the comparison failed due to incompatible units.
        """
        other_unit = getattr(other, "unit", 1.0)
        other_value = getattr(other, "value", other)

        if self.__unit == other_unit:
            return comparison_operator(other_value)

        alpha = other_unit / self.__unit

        if isinstance(alpha, float):
            return comparison_operator(other_value * alpha)

        if (
            isinstance(alpha, UnitExprBase)
            and alpha.base_exponents == alpha.base_exponents_zero
        ):
            return comparison_operator(other_value * alpha.base_factor)

        return None
